// ===========================================================================
// Filter.cpp
// color filter
//
//                                                       by kanegon 2006
// ===========================================================================

#include "stdafx.h"
#define _USE_MATH_DEFINES
#include <math.h>
#include "AnotherVision.h"
#include "common.h"
#include "profile.h"
#include "AppException.h"
#include "Filter.h"


// ---------------------------------------------------------------------------

#define BASE_COLOR_DIV       5
#define BASE_COLOR_DELTA     (255 / BASE_COLOR_DIV)
#define BASE_COLOR_MAX       (BASE_DELTA * BASE_COLOR_DIV)
#define BASE_COLOR_MAP_SIZE  (BASE_COLOR_DIV + 1) * (BASE_COLOR_DIV + 1) * (BASE_COLOR_DIV + 1);

// ---------------------------------------------------------------------------

static
int s_nColorExpand = 0;
static
int s_nColorDiv = 0;
static
int s_nColorElmSize = 0;
static
int s_nColorElmSize2 = 0;
static
int s_nColorMapSize = 0;

// ---------------------------------------------------------------------------
// 
// ---------------------------------------------------------------------------
void AppInitialiizeColorExpand(int nColorExpand)
{
	if (nColorExpand < 0 || nColorExpand > MAX_COLOR_EXTEND) {
		nColorExpand = 0;
	}
	s_nColorExpand = nColorExpand;
	s_nColorDiv = (nColorExpand == 0) ? 7 : nColorExpand;
	s_nColorElmSize = BASE_COLOR_DIV * s_nColorDiv + 1;
	s_nColorElmSize2 = s_nColorElmSize * s_nColorElmSize;
	s_nColorMapSize = s_nColorElmSize * s_nColorElmSize * s_nColorElmSize;
}

// ---------------------------------------------------------------------------
// 
// ---------------------------------------------------------------------------
int AppGetColorExpand()
{
	return s_nColorExpand;
}

// ---------------------------------------------------------------------------
// 
// ---------------------------------------------------------------------------
BOOL AppIsColorExpand()
{
	return (s_nColorExpand != 0);
}

// ---------------------------------------------------------------------------
// 
// ---------------------------------------------------------------------------
BOOL AppIsColorWordIndexes()
{
	return (s_nColorDiv <= 7);
}

// ---------------------------------------------------------------------------
// linear interpolation
// ---------------------------------------------------------------------------
inline
COLORREF GetInterpolationColor(COLORREF col1, COLORREF col2, int i, int n)
{
	int n1 = n - i;
	int n2 = i;
	return RGB(
			(GetRValue(col1) * n1 + GetRValue(col2) * n2) / n,
			(GetGValue(col1) * n1 + GetGValue(col2) * n2) / n,
			(GetBValue(col1) * n1 + GetBValue(col2) * n2) / n);
}

// ---------------------------------------------------------------------------

static
CFilterManager s_ColorFilterManager;


// ---------------------------------------------------------------------------

static
int GetColormapIndex(COLORREF clr);
static
int GetColormapExIndex(COLORREF clr);


// ---------------------------------------------------------------------------
// string to COLORREF
// string format must be "#[A-Fa-f0-9]{6}"
// ---------------------------------------------------------------------------
static inline
BOOL TextToColorref(LPCTSTR pszText, COLORREF& color)
{
	if (pszText == NULL) {
		return FALSE;
	}
	if (lstrlen(pszText) != 7) {
		return FALSE;
	}
	if (pszText[0] != '#') {
		return FALSE;
	}
	int i;
	for (i = 1; i <= 6; i++) {
		if (! _istxdigit((_TUCHAR)pszText[i])) {
			return FALSE;
		}
	}
	int r, g, b;
	int result = _stscanf(pszText, _T("#%02x%02x%02x"), &r, &g, &b);
	if (result != 3) {
		return FALSE;
	}
	color = RGB(r, g, b);
	return TRUE;
}

// ---------------------------------------------------------------------------
// 
// ---------------------------------------------------------------------------
static inline
_bstr_t GetAttribute(MSXML2::IXMLDOMElementPtr pElement, LPCTSTR pszAttrName)
{
	MSXML2::IXMLDOMAttributePtr pAttrNode = pElement->getAttributeNode(_bstr_t(pszAttrName));
	if (pAttrNode == NULL) {
		return (BSTR)NULL;
	}
	return pAttrNode->value;
}

// ---------------------------------------------------------------------------
// get unique instance of color filter manager class
// ---------------------------------------------------------------------------
CFilterManager* GetAppFilterManager()
{
	return &s_ColorFilterManager;
}

// ---------------------------------------------------------------------------
// expand color map (216 color => N color)
//
//      N (color resolusion) = (5 * nExpandColor + 1) ** 3
//      memory (byte)        = N * 4
//
//   nExpandColor   color resolusion   memory(byte)
//   ----------------------------------------------
//         0          (46656)
//         1             216                864
//         2            1331               5324
//         3            4096              16384
//         4            9261              37044
//         5           17576              70304
//         6           29791             119164
//         7           46656             186624
//         8           68921             275684
//         9           97336             389344
//        10          132651             530604
//        11          175616             702464
//        12          226981             907924  max: limited by dialog
//        13          287496            1149984
//        14          357911            1431644
//        15          438976            1755904
//        16          531441            2125764
//        17          636056            2544224
//        18          753571            3014284  max: limited by profile
// ---------------------------------------------------------------------------
BOOL CFilter::CreateColorMapEx()
{
	if (m_pColorMapEx != NULL) {
		return TRUE;
	}

	m_pColorMapEx = new COLORREF[s_nColorMapSize];
	if (m_pColorMapEx == NULL) {
		return FALSE;
	}

	int nIndex = 0;
	int r, g, b;
	int nBaseIndex = 0;
	for (r = 0; r < s_nColorElmSize; r += s_nColorDiv) {
		for (g = 0; g < s_nColorElmSize; g += s_nColorDiv) {
			for (b = 0; b < s_nColorElmSize; b += s_nColorDiv) {
				nIndex = r * s_nColorElmSize2 + g * s_nColorElmSize + b;
				m_pColorMapEx[nIndex] = m_aColorMap[nBaseIndex];
				++nBaseIndex;
			}
		}
	}
	for (r = 0; r < s_nColorElmSize; r += s_nColorDiv) {
		for (g = 0; g < s_nColorElmSize; g += s_nColorDiv) {
			for (b = 0; b < s_nColorElmSize - 1; b += s_nColorDiv) {
				nIndex = r * s_nColorElmSize2 + g * s_nColorElmSize + b;
				int bb;
				for (bb = 1; bb < s_nColorDiv; ++bb) {
					m_pColorMapEx[nIndex + bb] = GetInterpolationColor(
							m_pColorMapEx[nIndex], m_pColorMapEx[nIndex + s_nColorDiv],
							bb, s_nColorDiv);
				}
			}
		}
	}
	for (r = 0; r < s_nColorElmSize; r += s_nColorDiv) {
		for (g = 0; g < s_nColorElmSize - 1; g += s_nColorDiv) {
			for (b = 0; b < s_nColorElmSize; ++b) {
				nIndex = r * s_nColorElmSize2 + g * s_nColorElmSize + b;
				int gg;
				for (gg = 1; gg < s_nColorDiv; ++gg) {
					m_pColorMapEx[nIndex + gg * s_nColorElmSize] = GetInterpolationColor(
							m_pColorMapEx[nIndex], m_pColorMapEx[nIndex + s_nColorDiv * s_nColorElmSize],
							gg, s_nColorDiv);
				}
			}
		}
	}
	for (r = 0; r < s_nColorElmSize - 1; r += s_nColorDiv) {
		for (g = 0; g < s_nColorElmSize; ++g) {
			for (b = 0; b < s_nColorElmSize; ++b) {
				nIndex = r * s_nColorElmSize2 + g * s_nColorElmSize + b;
				int rr;
				for (rr = 1; rr < s_nColorDiv; ++rr) {
					m_pColorMapEx[nIndex + rr * s_nColorElmSize2] = GetInterpolationColor(
							m_pColorMapEx[nIndex], m_pColorMapEx[nIndex + s_nColorDiv * s_nColorElmSize2],
							rr, s_nColorDiv);
				}
			}
		}
	}
	return TRUE;
}

// ---------------------------------------------------------------------------
// delete color map
// ---------------------------------------------------------------------------
void CFilter::DeleteColorMapEx()
{
	if (m_pColorMapEx != NULL) {
		delete[] m_pColorMapEx;
		m_pColorMapEx = NULL;
	}
}

// ---------------------------------------------------------------------------
// check filter id
// ---------------------------------------------------------------------------
BOOL CFilter::IsValidFilterID(LPCTSTR pszID)
{
	if (pszID == NULL) {
		return FALSE;
	}
	int i;
	for (i = 0; pszID[i] != _T('\0'); i++) {
		_TUCHAR ch = pszID[i];
		if (! _istalnum(ch) && ch != _T('_') && ch != _T('@')) {
			return FALSE;
		}
	}
	if (i == 0 || i > MAX_FILTER_ID) {
		return FALSE;
	}
	return TRUE;
}

// ---------------------------------------------------------------------------
// check filter name
// ---------------------------------------------------------------------------
BOOL CFilter::IsValidFilterName(LPCTSTR pszName)
{
	if (pszName == NULL) {
		return FALSE;
	}
	int len = lstrlen(pszName);
	if (len == 0 || len > MAX_FILTER_NAME) {
		return FALSE;
	}
	return TRUE;
}

// ---------------------------------------------------------------------------
// check type value
// ---------------------------------------------------------------------------
BOOL CFilter::IsValidType(LPCTSTR pszType)
{
	//TODO:
	return TRUE;
}

// ---------------------------------------------------------------------------
// check visible value
// ---------------------------------------------------------------------------
BOOL CFilter::IsValidVisible(LPCTSTR pszVisible)
{
	//TODO:
	return TRUE;
}

// ---------------------------------------------------------------------------
// check repeat value
// ---------------------------------------------------------------------------
BOOL CFilter::IsValidRepeat(LPCTSTR pszRepeat)
{
	//TODO:
	return TRUE;
}

// ---------------------------------------------------------------------------
// check timeout value
// ---------------------------------------------------------------------------
BOOL CFilter::IsValidTimeout(LPCTSTR pszTimeout)
{
	//TODO:
	return TRUE;
}

// ---------------------------------------------------------------------------
// check filter tag of xml
// ---------------------------------------------------------------------------
BOOL CFilter::IsValidFilterTags(MSXML2::IXMLDOMElementPtr& pFilterListElement)
{
	MSXML2::IXMLDOMNodeListPtr pChildNodeList;
	pChildNodeList = pFilterListElement->childNodes;
	if (pChildNodeList == NULL) {
		APPTHROW(_T("bad filter format. (failed to get FilterList child.)"));
	}
	MSXML2::IXMLDOMNodePtr pNode;
	for (;;) {
		pNode = pChildNodeList->nextNode();
		if (pNode == NULL) {
			break;
		}
		MSXML2::DOMNodeType NodeType;
		NodeType = pNode->nodeType;
		if (NodeType != MSXML2::NODE_ELEMENT) {
			continue;
		}
		MSXML2::IXMLDOMElementPtr pFilterElement = pNode;
		_bstr_t bstrTagName = pNode->nodeName;
		if (lstrcmp((LPCTSTR)bstrTagName, _T("Filter")) != 0) {
			APPTHROW(_T("bad filter format. (not Filter tag as FilterList child.)"));
		}
		_bstr_t bstrValue =  GetAttribute(pFilterElement, _T("ID"));
		if ((LPCTSTR)bstrValue == NULL) {
			APPTHROW(_T("bad filter format. (ID attribute not found.)"));
		}
		xstr_t strID = (LPCTSTR)bstrValue;
		bstrValue =  GetAttribute(pFilterElement, _T("Timeout"));
		if ((LPCTSTR)bstrValue == NULL) {
			APPTHROW(_T("bad filter format. (Timeout attribute not found.)"));
		}
		if (! IsValidTimeout((LPCTSTR)bstrValue)) {
			APPTHROW(_T("bad filter format. (Timeout attribute is bad.)"));
		}
		int nTimeout = StrToInt((LPCTSTR)bstrValue);
		CFilterRef ref = CFilterRef((LPCTSTR)strID, NULL, nTimeout);
		m_RefFilters.push_back(ref);
	}
	return TRUE;
}

// ---------------------------------------------------------------------------
// check filterlist tag of xml
// ---------------------------------------------------------------------------
BOOL CFilter::IsValidFilterListTag(MSXML2::IXMLDOMElementPtr& pRootElement)
{
	//HRESULT hResult;
	MSXML2::IXMLDOMElementPtr pFilterListElement = pRootElement->selectSingleNode(_bstr_t(L"FilterList"));
	if (pFilterListElement == NULL) {
		APPTHROW(_T("bad filter format. (FilterList tag not found.)"));
	}
	if (! IsValidFilterTags(pFilterListElement)) {
		return FALSE;
	}
	return TRUE;
}

// ---------------------------------------------------------------------------
// check color tag of xml
// ---------------------------------------------------------------------------
BOOL CFilter::IsValidColorTags(MSXML2::IXMLDOMElementPtr& pColorMapElement)
{
	MSXML2::IXMLDOMNodeListPtr pChildNodeList;
	pChildNodeList = pColorMapElement->childNodes;
	if (pChildNodeList == NULL) {
		APPTHROW(_T("bad filter format. (failed to get ColorMap child.)"));
	}
	int nColorCount = 0;
	MSXML2::IXMLDOMNodePtr pNode;
	for (;;) {
		pNode = pChildNodeList->nextNode();
		if (pNode == NULL) {
			break;
		}
		MSXML2::DOMNodeType NodeType;
		NodeType = pNode->nodeType;
		if (NodeType != MSXML2::NODE_ELEMENT) {
			continue;
		}
		MSXML2::IXMLDOMElementPtr pColorElement = pNode;
		_bstr_t bstrTagName = pNode->nodeName;
		if (lstrcmp((LPCTSTR)bstrTagName, _T("Color")) != 0) {
			APPTHROW(_T("bad filter format. (not Color tag as ColorMap child.)"));
		}
		_bstr_t bstrValue =  GetAttribute(pColorElement, _T("RGB"));
		if ((LPCTSTR)bstrValue == NULL) {
			APPTHROW(_T("bad filter format. (RGB attribute not found.)"));
		}
		COLORREF color;
		if (! TextToColorref((LPCTSTR)bstrValue, color)) {
			APPTHROW(_T("bad filter format. (RGB attribute is bad.)"));
		}
		m_aColorMap[nColorCount] = color;
		++nColorCount;
	}
	if (nColorCount != 216) {
		// bad file format
		APPTHROW(_T("bad filter format. (number of Color tag is not 216.)"));
	}
	return TRUE;
}

// ---------------------------------------------------------------------------
// check colormap tag of xml
// ---------------------------------------------------------------------------
BOOL CFilter::IsValidColorMapTag(MSXML2::IXMLDOMElementPtr& pRootElement)
{
	//HRESULT hResult;
	MSXML2::IXMLDOMElementPtr pColorMapElement = pRootElement->selectSingleNode(_bstr_t(L"ColorMap"));
	if (pColorMapElement == NULL) {
		APPTHROW(_T("bad filter format. (ColorMap tag not found.)"));
	}
	if (! IsValidColorTags(pColorMapElement)) {
		return FALSE;
	}
	return TRUE;
}

// ---------------------------------------------------------------------------
// check profile tag of xml
// ---------------------------------------------------------------------------
BOOL CFilter::IsValidProfileTag(MSXML2::IXMLDOMElementPtr& pRootElement)
{
	//HRESULT hResult;
	MSXML2::IXMLDOMElementPtr pProfileElement = pRootElement->selectSingleNode(_bstr_t(L"Profile"));
	if (pProfileElement == NULL) {
		APPTHROW(_T("bad filter format. (Profile rag not found.)"));
	}
	_bstr_t bstrValue =  GetAttribute(pProfileElement, _T("ID"));
	if ((LPCTSTR)bstrValue == NULL) {
		APPTHROW(_T("bad filter format. (ID attribute not found.)"));
	}
	if (! IsValidFilterID((LPCTSTR)bstrValue)) {
		APPTHROW(_T("bad filter format. (filter ID is bad.)"));
	}
	m_strID = (LPCTSTR)bstrValue;

	bstrValue =  GetAttribute(pProfileElement, _T("DisplayName"));
	if ((LPCTSTR)bstrValue == NULL) {
		APPTHROW(_T("bad filter format. (DisplayName attribute not found.)"));
	}
	// trim?
	if (! IsValidFilterName((LPCTSTR)bstrValue)) {
		APPTHROW(_T("bad filter format. (filter DisplayName is bad.)"));
	}
	m_strName = (LPCTSTR)bstrValue;

	bstrValue = GetAttribute(pProfileElement, _T("Type"));
	if ((LPCTSTR)bstrValue == NULL) {
		m_uType = FILTERTYPE_DEFINITION;
	}
	else {
		if (! IsValidType((LPCTSTR)bstrValue)) {
			APPTHROW(_T("bad filter format. (Type attribute is bad.)"));
		}
		if (lstrcmp((LPCTSTR)bstrValue, _T("reference")) == 0) {
			m_uType = FILTERTYPE_REFERENCE;
		}
		else /* if (lstrcmp((LPCTSTR)bstrValue, _T("definition")) == FILTERTYPE_DEFINITION) */ {
			m_uType = FILTERTYPE_DEFINITION;
		}
	}

	bstrValue = GetAttribute(pProfileElement, _T("Visible"));
	if ((LPCTSTR)bstrValue == NULL) {
		m_bVisible = TRUE;
	}
	else {
		if (! IsValidVisible((LPCTSTR)bstrValue)) {
			APPTHROW(_T("bad filter format. (Visible attribute is bad.)"));
		}
		if (lstrcmp((LPCTSTR)bstrValue, _T("no")) == 0) {
			m_bVisible = FALSE;
		}
		else /* if (lstrcmp((LPCTSTR)bstrValue, _T("definition")) == FILTERTYPE_DEFINITION) */ {
			m_bVisible = TRUE;
		}
	}

	bstrValue = GetAttribute(pProfileElement, _T("Repeat"));
	if ((LPCTSTR)bstrValue == NULL) {
		m_nRepeat = 0;
	}
	else {
		if (! IsValidRepeat((LPCTSTR)bstrValue)) {
			APPTHROW(_T("bad filter format. (Repeat attribute is bad.)"));
		}
		m_nRepeat = StrToInt((LPCTSTR)bstrValue);
	}

	return TRUE;
}

// ---------------------------------------------------------------------------
// check root tag of xml
// ---------------------------------------------------------------------------
BOOL CFilter::IsValidRootTag(MSXML2::IXMLDOMDocumentPtr& pXMLDoc)
{
	MSXML2::IXMLDOMElementPtr pRootElement;
	pRootElement = pXMLDoc->documentElement;
	if (pRootElement == NULL) {
		APPTHROW(_T("bad filter format. (root tag not found.)"));
	}
	_bstr_t bstrTagName = pRootElement->nodeName;
	if (lstrcmp((LPCTSTR)bstrTagName, _T("AnotherVisionFilter")) != 0) {
		APPTHROW(_T("bad filter format. (bad root tag.)"));
	}
	_bstr_t bstrVersion = GetAttribute(pRootElement, _T("Version"));;
	if ((LPCTSTR)bstrVersion == NULL) {
		APPTHROW(_T("bad filter format. (Version attribute not found.)"));
	}
	if (lstrcmp((LPCTSTR)bstrVersion, _T("1.0")) != 0) {
		APPTHROW(_T("bad filter format. (bad filter version.)"));
	}
	if (! IsValidProfileTag(pRootElement)) {
		return FALSE;
	}
	if (IsDefFilter()) {
		if (! IsValidColorMapTag(pRootElement)) {
			return FALSE;
		}
	}
	else /* if (IsRefFilter()) */ {
		if (! IsValidFilterListTag(pRootElement)) {
			return FALSE;
		}
	}
	return TRUE;
}

// ---------------------------------------------------------------------------
// check filter xml
// ---------------------------------------------------------------------------
BOOL CFilter::IsValidFilterXml(MSXML2::IXMLDOMDocumentPtr pXMLDoc)
{
	return IsValidRootTag(pXMLDoc);
}

// ---------------------------------------------------------------------------
// constructor
// ---------------------------------------------------------------------------
CFilter::CFilter()
{
	m_strID = _T("");
	m_strName = _T("");
	m_uType = FILTERTYPE_DEFINITION;
	m_bVisible = TRUE;
	m_bEnableInner = FALSE;
	m_bNormalFilter = FALSE;
	ZeroMemory(m_aColorMap, sizeof(m_aColorMap));
	m_pColorMapEx = NULL;
	m_nRepeat = 0;
	m_dwHotKey = 0;
}

// ---------------------------------------------------------------------------
// constructor
// ---------------------------------------------------------------------------
CFilter::CFilter(LPCTSTR pszID, LPCTSTR pszName, COLORREF* pColorMap)
{
	m_strID = pszID;
	m_strName = pszName;
	m_uType = FILTERTYPE_DEFINITION;
	m_bVisible = TRUE;
	m_bEnableInner = FALSE;
	m_bNormalFilter = FALSE;
	memcpy(m_aColorMap, pColorMap, sizeof(m_aColorMap));
	m_pColorMapEx = NULL;
	m_nRepeat = 0;
	m_dwHotKey = 0;
}

// ---------------------------------------------------------------------------
// constructor
// ---------------------------------------------------------------------------
CFilter::CFilter(const CFilter& filter)
{
	*this = filter;
}

// ---------------------------------------------------------------------------
// destructor
// ---------------------------------------------------------------------------
CFilter::~CFilter()
{
}

// ---------------------------------------------------------------------------
// operator=
// ---------------------------------------------------------------------------
CFilter& CFilter::operator=(const CFilter& filter)
{
	m_strID = filter.m_strID;
	m_strName = filter.m_strName;
	m_uType = filter.m_uType;
	m_bVisible = filter.m_bVisible;
	m_bEnableInner = filter.m_bEnableInner;
	m_bNormalFilter = filter.m_bNormalFilter;
	memcpy(m_aColorMap, filter.m_aColorMap, sizeof(m_aColorMap));
	if (filter.m_pColorMapEx != NULL) {
		m_pColorMapEx = new COLORREF[s_nColorMapSize];
		memcpy(m_pColorMapEx, filter.m_pColorMapEx, sizeof(s_nColorMapSize) * sizeof(COLORREF));
	}
	else {
		m_pColorMapEx = NULL;
	}
	int i;
	for (i = 0; i < (int)filter.m_RefFilters.size(); i++) {
		CFilterRef ref = filter.m_RefFilters[i];
		m_RefFilters.push_back(ref);
	}

	m_nRepeat = filter.m_nRepeat;
	m_dwHotKey = filter.m_dwHotKey;
	return *this;
}

#if 0
// ---------------------------------------------------------------------------
// convert color (from COLORREF) for 216 color
// ---------------------------------------------------------------------------
COLORREF CFilter::ConvertColor(COLORREF clr)
{
	COLORREF colorNew = m_aColorMap[GetColormapIndex(clr)];
	return colorNew;
}

// ---------------------------------------------------------------------------
// convert color (from index) for 216 color
// ---------------------------------------------------------------------------
COLORREF CFilter::ConvertColorFromIndex(int nColorIndex)
{
	COLORREF colorNew = m_aColorMap[nColorIndex];
	return colorNew;
}

// ---------------------------------------------------------------------------
// get ColorMap table for 216 color
// ---------------------------------------------------------------------------
COLORREF* CFilter::GetColorMapTable()
{
	return &m_aColorMap[0];
}
#endif

// ---------------------------------------------------------------------------
// convert color (from COLORREF) for N color
// ---------------------------------------------------------------------------
COLORREF CFilter::ConvertColorEx(COLORREF clr)
{
	COLORREF colorNew = m_pColorMapEx[GetColormapExIndex(clr)];
	return colorNew;
}

// ---------------------------------------------------------------------------
// convert color (from index) for N color
// ---------------------------------------------------------------------------
COLORREF CFilter::ConvertColorExFromIndex(int nColorIndex)
{
	COLORREF colorNew = m_pColorMapEx[nColorIndex];
	return colorNew;
}

// ---------------------------------------------------------------------------
// get ColorMap table for N color
// ---------------------------------------------------------------------------
COLORREF* CFilter::GetColorMapExTable()
{
	return m_pColorMapEx;
}

// ---------------------------------------------------------------------------
// read filter xml file
// ---------------------------------------------------------------------------
BOOL CFilter::ReadFile(LPCTSTR pszPath)
{
	HRESULT hResult;
	try {
		try {
			if (! IsFile(pszPath)) {
				APPTHROW(_T("file not found."));
			}
			MSXML2::IXMLDOMDocumentPtr pXMLDoc;
			hResult = ::CoCreateInstance(
					MSXML2::CLSID_DOMDocument,
					NULL,
					CLSCTX_INPROC_SERVER,
					MSXML2::IID_IXMLDOMDocument,
					(void**)&pXMLDoc);
			ValidateHResult(hResult);
			pXMLDoc->async = VARIANT_FALSE;
			_variant_t vFilePath = _bstr_t(pszPath);
			VARIANT_BOOL vResult;
			vResult = pXMLDoc->load(vFilePath);
			if (vResult != VARIANT_TRUE) {
				APPTHROW(_T("xml parse error."));
			}
			BOOL bResult = IsValidFilterXml(pXMLDoc);
			return bResult;
		}
		catch (_com_error& e) {
			APPTHROW_COM_ERROR(e);
		}
		catch (exception& e) {
			APPTHROW_EXCEPTION(e);
		}
	}
	catch (AppException& e) {
		ATLTRACE2(atlTraceGeneral, 0, _T("%s\n"), e.GetMessage());
	}
	return FALSE;
}

#if 0
// ---------------------------------------------------------------------------
// color => index convert
// ---------------------------------------------------------------------------
// expression:
//   (n + 25) div 51
// mapping:
//   0:   0- 25  => 0
//   1:  26- 76  => 51
//   2:  77-127  => 102
//   3: 128-178  => 153
//   4: 179-229  => 204
//   5: 230-255  => 255
// ---------------------------------------------------------------------------
int CFilter::GetColormapIndex(COLORREF clr)
{
	return
		((GetRValue(clr) + 25) / 51) * 36 +
		((GetGValue(clr) + 25) / 51) * 6 +
		((GetBValue(clr) + 25) / 51);
}
#endif

// ---------------------------------------------------------------------------
// color => index convert (N color version)
// ---------------------------------------------------------------------------
int CFilter::GetColormapExIndex(COLORREF clr)
{
	return
		(((GetRValue(clr) * s_nColorDiv) + 25) / 51) * s_nColorElmSize2 +
		(((GetGValue(clr) * s_nColorDiv) + 25) / 51) * s_nColorElmSize +
		(((GetBValue(clr) * s_nColorDiv) + 25) / 51);
}

// ---------------------------------------------------------------------------
// create instance for normal filter
// ---------------------------------------------------------------------------
CFilter* CFilter::CreateNormalFileter()
{
	COLORREF aColorMap[216];
	int nIndex = 0;
	int r, g, b;
	for (r = 0; r <= 255; r += 51) {
		for (g = 0; g <= 255; g += 51) {
			for (b = 0; b <= 255; b += 51) {
				aColorMap[nIndex++] = RGB(r, g, b);
			}
		}
	}
	CFilter* pFilter = new CFilter(FILTER_ID_NORMAL, FILTER_INITIALIZE_NAME_NORMAL, aColorMap);
	pFilter->m_bNormalFilter = TRUE;
	return pFilter;
}

// ---------------------------------------------------------------------------
// 
// ---------------------------------------------------------------------------
int CFilter::GetRefFilterCount()
{
	return (int)m_RefFilters.size();
}

// ---------------------------------------------------------------------------
// 
// ---------------------------------------------------------------------------
void CFilter::SetRepeat(int nRepeat)
{
	m_nRepeat = nRepeat;
}

// ---------------------------------------------------------------------------
// 
// ---------------------------------------------------------------------------
int CFilter::GetRepeat()
{
	return m_nRepeat;
}

// ---------------------------------------------------------------------------
// 
// ---------------------------------------------------------------------------
BOOL CFilter::GetFilterRef(int nIndex, CFilterRef& ref)
{
	if (! IsRefFilter()) {
		return FALSE;
	}
	if (nIndex < 0 || nIndex >= (int)m_RefFilters.size()) {
		return FALSE;
	}
	ref = m_RefFilters[nIndex];
	return TRUE;
}

// ---------------------------------------------------------------------------
// 
// ---------------------------------------------------------------------------
BOOL CFilter::IsValidateRefFilter()
{
	if (! IsRefFilter()) {
		return FALSE;
	}
	int i;
	for (i = 0; i < (int)m_RefFilters.size(); i++) {
		CFilterRef& ref = m_RefFilters[i];
		CFilter* pFilter = GetAppFilterManager()->GetFilter((LPCTSTR)ref.GetID());
		if (pFilter == NULL) {
			return FALSE;
		}
		ref.SetFilter(pFilter);
	}
	return TRUE;
}

// ---------------------------------------------------------------------------
// initialize
// ---------------------------------------------------------------------------
void CFilterManager::Initialize()
{
	CFilter* pFilter = CFilter::CreateNormalFileter();
	Register(pFilter);
}

// ---------------------------------------------------------------------------
// constructor
// ---------------------------------------------------------------------------
CFilterManager::CFilterManager()
{
	Initialize();
}

// ---------------------------------------------------------------------------
// destructor
// ---------------------------------------------------------------------------
CFilterManager::~CFilterManager()
{
	int i;
	for (i = 0; i < (int)m_Filters.size(); i++) {
		CFilter* pFilter = m_Filters[i];
		if (pFilter != NULL) {
			delete pFilter;
		}
	}
	m_Filters.clear();
}

// ---------------------------------------------------------------------------
// read filter files
// ---------------------------------------------------------------------------
BOOL CFilterManager::ReadFilters(LPCTSTR pszFilterDir)
{
	if (! IsDirectory(pszFilterDir)) {
		return FALSE;
	}

	TCHAR szSearchPath[_MAX_PATH];
	lstrcpy(szSearchPath, pszFilterDir);
	PathAppend(szSearchPath, APP_FILTER_FILE);

	COLOR_FILTER_ARRAY RefFilters;

	WIN32_FIND_DATA find;
	HANDLE hFind = FindFirstFile(szSearchPath, &find);
	if (hFind == INVALID_HANDLE_VALUE) {
		return FALSE;
	}
	do {
		if (! IsFlag(find.dwFileAttributes, FILE_ATTRIBUTE_DIRECTORY)) {
			TCHAR szPath[_MAX_PATH];
			lstrcpy(szPath, pszFilterDir);
			PathAppend(szPath, find.cFileName);
			CFilter filter;
			if (filter.ReadFile(szPath)) {
				if (filter.IsDefFilter()) {
					CFilter* pFilter = GetFilter(filter.GetID());
					if (pFilter == NULL) {
						Register(new CFilter(filter));
					}
					else if (lstrcmpi(filter.GetID(), FILTER_ID_NORMAL) == 0 &&
							lstrcmp(pFilter->GetName(), FILTER_INITIALIZE_NAME_NORMAL) == 0) {
						pFilter->SetName(filter.GetName());
					}
				}
				else {
					RefFilters.push_back(filter);
				}
			}
			else {
				ATLTRACE2(atlTraceGeneral, 0, _T("CFilter::ReadFile() failed. [file=%s]\n"), szPath);
			}
		}
	} while(FindNextFile(hFind, &find) != 0);
	FindClose(hFind);

	int i;
	for (i = 0; i < (int)RefFilters.size(); i++) {
		CFilter& filter = RefFilters[i];
		if (filter.IsValidateRefFilter()) {
			if (GetFilter(filter.GetID()) == NULL) {
				Register(new CFilter(filter));
			}
		}
	}

	return TRUE;
}
